/*  
   $Id: tcap.c,v 1.5 2005/01/16 03:26:06 jose Exp $ 

   minimal Tcl pcap interface (tcap), 
   loosely based on http://wiki.tcl.tk/10877 
 */

#define  TCAP_VERSION        "0.1.2"

#include <pcap.h>
#include <stdlib.h>
#include <string.h>

#include <tcl.h>

#include "pcaputil.h"

static pcap_t	*pcap = NULL;
static char	 helpstr[] = 
	"tcap - Tcl pcap(3) interface\n"
	"options:\n"
        "       list\n"
        "         list all available devices.\n"
	"	open dev promisc snaplen\n"
	"	  opens a new packet capture interface.\n"
	"	dloff\n"
	"	  returns the datalink offset in byes.\n"
	"	filter bpg_program\n"
	"	  sets a bpf(3) filter on the pcap interface.\n"
	"	get\n"
	"	  returns the next packet as a list of bytes.\n"
	"	close\n"
	"	  closes the capture interface.";

static int 	tcapCmd _ANSI_ARGS_ ((ClientData dummy,
                                Tcl_Interp *interp, int objc,
                                Tcl_Obj *CONST objv[]));
static void 	tcap_exit(ClientData dummy); /* exit handler */

#define MAX_ERROR_SIZE   1024

static int
setTclError TCL_VARARGS_DEF (Tcl_Interp *, i)
{
     	va_list argList;
     	char buf[MAX_ERROR_SIZE];
     	char *format;

     	Tcl_Interp *interp = TCL_VARARGS_START(Tcl_Interp *, i, argList);
     	format = va_arg(argList, char *);
     	vsnprintf(buf, MAX_ERROR_SIZE, format, argList);
     	buf[MAX_ERROR_SIZE-1] = '\0';
     	Tcl_SetResult(interp, buf, TCL_VOLATILE);
	return TCL_ERROR;
}

int
Tcap_Init(Tcl_Interp *interp)
{
	if (Tcl_PkgProvide(interp, "tcap", TCAP_VERSION) == TCL_ERROR) 
		return TCL_ERROR;
	Tcl_CreateObjCommand(interp, "tcap", tcapCmd, (ClientData) NULL, 
			    (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateExitHandler(tcap_exit,(ClientData) NULL);
	return TCL_OK;
}

#ifdef __APPLE__	/* for OS X */
int
Tcap_SafeInit(Tcl_Interp *interp)
{
	return TCL_OK;
}
#endif /* __APPLE__ */

static void
tcap_exit(ClientData dummy)
{
	if (pcap != NULL)
		pcap_close(pcap);
}

static int
tcapCmd(ClientData dummy, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
	int index;

	static CONST char *optionStrings[] = {
		"open", "list", "dloff", "filter", "get", "close", "help",
		NULL
	};

	enum options {
		TCAP_OPEN, TCAP_LIST, TCAP_DLOFF, TCAP_FILTER, TCAP_GET, 
		TCAP_CLOSE, TCAP_HELP
	};

	if (Tcl_GetIndexFromObj(interp, objv[1], optionStrings, "option", 0,
			        &index) != TCL_OK)
		return TCL_ERROR;

	switch ((enum options) index) {
	case TCAP_OPEN: {
		int promisc, snaplen;
		char *name;

		if (objc != 5) {
			Tcl_WrongNumArgs(interp, 2, objv, "dev promisc snaplen");
			return TCL_ERROR;
		}
		name = Tcl_GetStringFromObj(objv[2], NULL);
		if (Tcl_GetIntFromObj(interp, objv[3], &promisc) != TCL_OK)
			return TCL_ERROR;
		if (Tcl_GetIntFromObj(interp, objv[4], &snaplen) != TCL_OK)
			return TCL_ERROR;

		pcap = pcap_open(name, promisc, snaplen);
		if (pcap == NULL)
        		return setTclError(interp, "error opening %s", name);
		return TCL_OK;
	}	/* TCAP_OPEN */

	case TCAP_LIST: {
	       char *devlist;
	       pcap_if_t *alldevs, *d;
	       char errbuf[PCAP_ERRBUF_SIZE];

	       devlist = malloc(1024);
	       if (devlist == NULL)
		      return TCL_ERROR;
	       bzero(devlist, 1024);

	       if (pcap_findalldevs(&alldevs, errbuf) == -1) {
		       return TCL_ERROR;
	       }
	       for (d = alldevs; d; d = d->next) {
		       snprintf(devlist, 1024, "%s%s ", devlist, d->name);
	       }
	       pcap_freealldevs(alldevs);
	       Tcl_SetResult(interp, devlist, TCL_VOLATILE);
	       return TCL_OK;
        }       /* TCAP_LIST */

	case TCAP_DLOFF: {
		int dloff;
		char off[3];

		if (pcap == NULL)
			return TCL_ERROR;
		dloff = pcap_dloff(pcap);
		if (dloff < 0)
        		return setTclError(interp, "no opened handle");
		snprintf(off, sizeof(off), "%d", dloff);
		Tcl_SetResult(interp, off, TCL_VOLATILE); 
		return TCL_OK;
	} 	/* TCAP_DLOFF */

	case TCAP_FILTER: {
		char *filter;

		if (objc != 3) {
			Tcl_WrongNumArgs(interp, 2, objv, "bpf_program");
			return TCL_ERROR;	
		}
		if (pcap == NULL)
			return TCL_ERROR;
		filter = Tcl_GetStringFromObj(objv[2], NULL);
		if (pcap_filter(pcap, filter) < 0) 
        		return setTclError(interp, "error in filter");
		return TCL_OK;	
	}	/* TCAP_FILTER */

	case TCAP_GET: {
		u_char *pkt;
		struct pcap_pkthdr ph;
		int i, s;
		char *res;

		if (pcap == NULL)
			return TCL_ERROR;
		pkt = (u_char *)pcap_next(pcap, &ph);
		if (pkt == NULL)
			return TCL_OK;
		s = ph.len;
		res = malloc(s);
		if (res == NULL)
			return TCL_ERROR;
		bzero(res, s);
		for (i = 0; i < s; i++) 
			snprintf(res, s, "%s%02x ", res, pkt[i]);
		Tcl_SetResult(interp, res, TCL_STATIC);
		return TCL_OK;
	}	/* TCAP_GET */

	case TCAP_CLOSE: {
		if (pcap == NULL)
			return TCL_ERROR;
		pcap_close(pcap);
		return TCL_OK;
	}	/* TCAP_CLOSE */

	case TCAP_HELP: {
		Tcl_SetResult(interp, helpstr, TCL_VOLATILE);
		return TCL_OK;
	} 	/* TCAP_HELP */

	default: {
        	return setTclError(interp, "Couldn't understand arg %d as "
                            "action type", index);
     	}	/* default */
	}
}
